import "@moonbeam-network/api-augment";
import { expect } from "chai";
import { describeDevMoonbeamAllEthTxTypes } from "../../util/setup-dev-tests";
import { ethers } from "ethers";
import { getCompiled } from "../../util/contracts";
import {
  ALITH_TRANSACTION_TEMPLATE,
  createContract,
  createTransaction,
} from "../../util/transactions";
import { u8aToHex } from "@polkadot/util";

import { DEFAULT_GENESIS_MAPPING, PRECOMPILE_AUTHOR_MAPPING_ADDRESS } from "../../util/constants";
import { alith, generateKeyingPair } from "../../util/accounts";
import { KeyringPair } from "@substrate/txwrapper-core";

const AUTHOR_MAPPING_CONTRACT = getCompiled("AuthorMapping");
const AUTHOR_MAPPING_INTERFACE = new ethers.utils.Interface(AUTHOR_MAPPING_CONTRACT.contract.abi);

describeDevMoonbeamAllEthTxTypes("Precompiles - author mapping", (context) => {
  it("allows to add association", async function () {
    const mappingAccount = generateKeyingPair("sr25519");
    const { rawTx } = await createContract(context, "AuthorMapping");
    await context.createBlock(rawTx);

    const { result } = await context.createBlock(
      createTransaction(context, {
        ...ALITH_TRANSACTION_TEMPLATE,
        to: PRECOMPILE_AUTHOR_MAPPING_ADDRESS,
        data: AUTHOR_MAPPING_INTERFACE.encodeFunctionData("add_association", [
          mappingAccount.publicKey,
        ]),
      })
    );

    const receipt = await context.web3.eth.getTransactionReceipt(result.hash);
    expect(receipt.status).to.equal(true);

    const mapping = await context.polkadotApi.query.authorMapping.mappingWithDeposit(
      u8aToHex(mappingAccount.publicKey)
    );
    expect(mapping.unwrap().account.toString()).to.eq(alith.address);
    expect(mapping.unwrap().deposit.toBigInt()).to.eq(DEFAULT_GENESIS_MAPPING);
  });
});

describeDevMoonbeamAllEthTxTypes("Precompiles - author mapping", (context) => {
  let firstMappingAccount: KeyringPair;
  let secondMappingAccount: KeyringPair;
  before("First add association", async () => {
    firstMappingAccount = generateKeyingPair("sr25519");
    secondMappingAccount = generateKeyingPair("sr25519");
    // Add association
    await context.createBlock(
      context.polkadotApi.tx.authorMapping.addAssociation(firstMappingAccount.publicKey)
    );

    // Verify association was added
    const mapping = await context.polkadotApi.query.authorMapping.mappingWithDeposit(
      u8aToHex(firstMappingAccount.publicKey)
    );
    expect(mapping.unwrap().account.toString()).to.eq(alith.address);
    expect(mapping.unwrap().deposit.toBigInt()).to.eq(DEFAULT_GENESIS_MAPPING);
  });

  it("allows to update association", async function () {
    const { rawTx } = await createContract(context, "AuthorMapping");
    await context.createBlock(rawTx);

    const tx = await createTransaction(context, {
      to: PRECOMPILE_AUTHOR_MAPPING_ADDRESS,
      data: AUTHOR_MAPPING_INTERFACE.encodeFunctionData("update_association", [
        firstMappingAccount.publicKey,
        secondMappingAccount.publicKey,
      ]),
    });

    const { result } = await context.createBlock(tx);
    const receipt = await context.web3.eth.getTransactionReceipt(result.hash);
    expect(receipt.status).to.equal(true);

    // Verify we updated firstMappingAccount for secondMappingAccount
    const mapping = await context.polkadotApi.query.authorMapping.mappingWithDeposit(
      u8aToHex(secondMappingAccount.publicKey)
    );
    expect(mapping.unwrap().account.toString()).to.eq(alith.address);
    expect(mapping.unwrap().deposit.toBigInt()).to.eq(DEFAULT_GENESIS_MAPPING);
    expect(
      (
        await context.polkadotApi.query.authorMapping.mappingWithDeposit(
          u8aToHex(firstMappingAccount.publicKey)
        )
      ).isNone
    ).to.be.true;
  });
});

describeDevMoonbeamAllEthTxTypes("Precompiles - author mapping", (context) => {
  let mappingAccount: KeyringPair;
  before("First add association", async () => {
    mappingAccount = generateKeyingPair("sr25519");
    // Add association

    await context.createBlock(
      context.polkadotApi.tx.authorMapping.addAssociation(mappingAccount.publicKey)
    );

    // Verify association was added
    const mapping = await context.polkadotApi.query.authorMapping.mappingWithDeposit(
      u8aToHex(mappingAccount.publicKey)
    );
    expect(mapping.unwrap().account.toString()).to.eq(alith.address);
    expect(mapping.unwrap().deposit.toBigInt()).to.eq(DEFAULT_GENESIS_MAPPING);
  });
  it("allows to clear association", async function () {
    const { rawTx } = await createContract(context, "AuthorMapping");
    await context.createBlock(rawTx);

    const { result } = await context.createBlock(
      createTransaction(context, {
        ...ALITH_TRANSACTION_TEMPLATE,
        to: PRECOMPILE_AUTHOR_MAPPING_ADDRESS,
        data: AUTHOR_MAPPING_INTERFACE.encodeFunctionData("clear_association", [
          mappingAccount.publicKey,
        ]),
      })
    );

    const receipt = await context.web3.eth.getTransactionReceipt(result.hash);
    expect(receipt.status).to.equal(true);

    // Verify we removed the association
    expect(
      (
        await context.polkadotApi.query.authorMapping.mappingWithDeposit(
          u8aToHex(mappingAccount.publicKey)
        )
      ).isNone
    ).to.be.true;
  });
});
